/** \file   dd_hit.c
 *  \brief  Implementation of hit operation (and minhit operation by Knuth)
 *  \author Takahisa Toda
 *  \note See: T.Toda, Hypergraph Transversal Computation with Binary Decision Diagrams, in Proc. of 12th International Symposium on Experimental Algorithms (SEA2013), LNCS 7933, pp.91-102, Rome, Italy (2013). 
 */

#include "my_def.h"
#include "my_hash.h"
#include "zdd_extra.h"
#include "dd_wrapper.h"
#include "dd_hit.h"

static my_bddp hit_z2b_rec(my_zddp f, struct htable *h);
static my_zddp minhit_nonsup_rec(my_zddp f, struct htable *h);

#ifdef SIZE_LOG_EXP
static size_t maxbddsize = 0; //!< maximum size of BDDs
#endif /*SIZE_LOG_EXP*/
 
static size_t recdepth = 0; //!< recursion depth

/** \brief Construct the BDD that represent all hitting sets for a ZDD.
 * \param f ZDD
 * \return the contructed BDD.
 * \see hit_z2b_rec
 */
my_bddp hit_z2b(my_zddp f)
{
  struct htable *h = ht_create(HASH_DEFAULT_SIZE); /*adjust HASH_DEFAULT_SIZE if necessary.*/
  ENSURE_TRUE_MSG(h != NULL, "hash table creation failed");

#ifdef SIZE_LOG_EXP
  maxbddsize = 0;
#endif /*SIZE_LOG_EXP*/

  my_bddp r = hit_z2b_rec(f, h);
  ENSURE_TRUE(r != BDD_NULL);

#ifdef SIZE_LOG_EXP
  printf("maxbddsize:\t%zu\n", maxbddsize);
#endif /*SIZE_LOG_EXP*/
#if defined(LOG_LEVEL) && LOG_LEVEL <= LOG_DEBUG
  printf("maxhashdepth\t%zu\n", ht_maxdepth(h));
#endif /*defined(LOG_LEVEL) && LOG_LEVEL <= LOG_DEBUG*/

  ht_destroy(h); 
  return r;
}


static my_bddp hit_z2b_rec(my_zddp f, struct htable *h)
{
  if(f == zdd_top()) return bdd_bot();
  if(f == zdd_bot()) return bdd_top();

  int res; my_bddp r;
  if((res = ht_search((size_t)f, (size_t*)&r, h)) == MY_SUCCESS) return r;

  INC_RECDEPTH(recdepth);  
  my_bddp rl = hit_z2b_rec(zdd_LO(f), h);
  my_bddp rh = hit_z2b_rec(zdd_HI(f), h);
  my_bddp t = bdd_unique(zdd_label(f), rh, bdd_top());
  ENSURE_TRUE_MSG(t != BDD_NULL, "BDD operation failed");
  r = bdd_and(rl, t);
  ENSURE_TRUE_MSG(r != BDD_NULL, "BDD operation failed");

  ht_add((size_t)f, (size_t)r, h);

#ifdef SIZE_LOG_EXP
  const size_t outsize = bdd_size(r);
  if(maxbddsize < outsize) maxbddsize = outsize;
#endif /*SIZE_LOG_EXP*/

  DEC_RECDEPTH(recdepth);
  return r;
}



/** \brief Construct the ZDD that represents all minimal hitting sets for a ZDD.
 * \param f ZDD
 * \return the contructed ZDD.
 * \see minhit_nonsup_rec, nonsup
 * \note
 * - See pp.669-670 in "The Art of Computer Programming, Volume 4a (2011)" by Knuth
 */
my_zddp minhit_nonsup(my_zddp f)
{
  struct htable *h = ht_create(HASH_DEFAULT_SIZE); /*adjust HASH_DEFAULT_SIZE if necessary.*/
  ENSURE_TRUE_MSG(h != NULL, "hash table creation failed");

  my_zddp r = minhit_nonsup_rec(f, h);
  ENSURE_TRUE(r != ZDD_NULL);

#if defined(LOG_LEVEL) && LOG_LEVEL <= LOG_DEBUG
  printf("maxhashdepth\t%zu\n", ht_maxdepth(h));
#endif /*defined(LOG_LEVEL) && LOG_LEVEL <= LOG_DEBUG*/

  ht_destroy(h);
  return r;
}


static my_zddp minhit_nonsup_rec(my_zddp f, struct htable *h)
{
  if(f == zdd_bot()) return zdd_top();
  if(f == zdd_top()) return zdd_bot();

  int res; my_zddp r; 
  if((res = ht_search((size_t)f, (size_t*)&r, h)) == MY_SUCCESS) return r;

  INC_RECDEPTH(recdepth);  
  my_zddp t  = zdd_union(zdd_LO(f), zdd_HI(f));
  ENSURE_TRUE_MSG(t != ZDD_NULL, "ZDD operation failed");
  my_zddp rl = minhit_nonsup_rec(t, h);
  my_zddp tt  = minhit_nonsup_rec(zdd_LO(f), h);
  my_zddp rh = nonsup(tt, rl);
  ENSURE_TRUE_MSG(rh != ZDD_NULL, "ZDD extra operation failed");
  r = zdd_unique(zdd_label(f), rl, rh);
  ENSURE_TRUE_MSG(r != ZDD_NULL, "ZDD operation failed");

  ht_add((size_t)f, (size_t)r, h);
  DEC_RECDEPTH(recdepth);
  return r;
}


