/** \file   dd_min.c
 *  \brief  Implementation of min operation
 *  \author Takahisa Toda
 */

#include "my_def.h"
#include "my_hash.h"
#include "dd_wrapper.h"
#include "dd_min.h"

static my_zddp minimal_b2z_monotone_rec(my_bddp f, struct htable *h);

#ifdef SIZE_LOG_EXP
static size_t maxzddsize = 0; //!< maximum size of ZDDs
#endif /*SIZE_LOG_EXP*/

static size_t recdepth = 0; //!< recursion depth

/** \brief Construct the ZDD that represents all minimal sets in a BDD.
 * \param f BDD
 * \return the contructed ZDD.
 * \see minimal_b2z_monotone_rec
 * \note
 * - the Boolean function represented by an input BDD must be monotone.
 * - See pp.255-256 in "The Art of Computer Programming, Volume 4a (2011)" by Knuth
 */
my_zddp minimal_b2z_monotone(my_bddp f)
{
  struct htable *h = ht_create(HASH_DEFAULT_SIZE); /*adjust HASH_DEFAULT_SIZE if necessary.*/
  ENSURE_TRUE_MSG(h != NULL, "hash table creation failed");

#ifdef SIZE_LOG_EXP
  maxzddsize = 0;
#endif /*SIZE_LOG_EXP*/

  my_zddp r = minimal_b2z_monotone_rec(f, h);
  ENSURE_TRUE(r != ZDD_NULL);

#ifdef SIZE_LOG_EXP
  printf("maxzddsize:\t%zu\n", maxzddsize);
#endif /*SIZE_LOG_EXP*/
#if defined(LOG_LEVEL) && LOG_LEVEL <= LOG_DEBUG
  printf("maxhashdepth\t%zu\n", ht_maxdepth(h));
#endif /*defined(LOG_LEVEL) && LOG_LEVEL <= LOG_DEBUG*/

  ht_destroy(h); 
  return r;
}


static my_zddp minimal_b2z_monotone_rec(my_bddp f, struct htable *h)
{
  if(f == bdd_bot()) return zdd_bot();
  if(f == bdd_top()) return zdd_top();

  int res; my_zddp r;
  if((res = ht_search((size_t)f, (size_t*)&r, h)) == MY_SUCCESS) return r;

  INC_RECDEPTH(recdepth);  
  my_zddp mh = minimal_b2z_monotone_rec(bdd_HI(f), h);
  my_zddp ml = minimal_b2z_monotone_rec(bdd_LO(f), h);
  my_zddp t = zdd_diff(mh, ml);
  ENSURE_TRUE_MSG(t != ZDD_NULL, "ZDD operation failed");
  r = zdd_unique(bdd_label(f), ml, t);
  ENSURE_TRUE_MSG(r != ZDD_NULL, "ZDD operation failed");

  ht_add((size_t)f, (size_t)r, h);

#ifdef SIZE_LOG_EXP
  const size_t outsize = zdd_size(r);
  if(maxzddsize < outsize) maxzddsize = outsize;
#endif /*SIZE_LOG_EXP*/

  DEC_RECDEPTH(recdepth);
  return r;
}

