/** \file   dd_wrapper.h
 *  \brief  Functions to access BDD/ZDD libraries.
 *  \author Takahisa Toda
 */
#ifndef DD_WRAPPER_H
#define DD_WRAPPER_H

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <limits.h>

#include "my_def.h"

#if defined(SELECT_CUDD)
/*---------------------------------------------------------------------------*/
/* COLORADO BDD/ZDD PACKAGE                                                  */
/*---------------------------------------------------------------------------*/
#include "util.h"
#include "cudd.h"
#include "cuddInt.h"

#define BDD_PACKAGE "CU Decision Diagram Package Release 2.5.0"
#define ZDD_PACKAGE "CU Decision Diagram Package Release 2.5.0"
#define BDD_NULL  NULL
#define ZDD_NULL  NULL
#define BDD_VAR_MAXINDEX  CUDD_MAXINDEX //!< maximum value that a BDD library can handle.
#define ZDD_VAR_MAXINDEX  CUDD_MAXINDEX //!< maximum value that a ZDD library can handle.

typedef DdNode *my_bddp; //!< pointer to a BDD node
typedef DdNode *my_zddp; //!< pointer to a ZDD node

extern DdManager *dd_mgr;

/* Operations for BDDs */
static inline int bdd_init(size_t maxval, size_t cach_size)
{
  ENSURE_TRUE_MSG(maxval != 0 && maxval <= BDD_VAR_MAXINDEX && maxval < INFTY, "the number of variables is invalid.");
  if(dd_mgr == NULL) {
    dd_mgr = Cudd_Init((DdHalfWord)maxval, (DdHalfWord)maxval, CUDD_UNIQUE_SLOTS, 16777216, 8589934592);
    //dd_mgr = Cudd_Init((DdHalfWord)maxval, (DdHalfWord)maxval, CUDD_UNIQUE_SLOTS, CUDD_CACHE_SLOTS, 0);
    ENSURE_TRUE_MSG(dd_mgr != NULL, "BDD manager initialization failed.\n");
    Cudd_DisableGarbageCollection(dd_mgr);// disable GC since we never do dereference of nodes.
#if defined(LOG_LEVEL) && LOG_LEVEL <= LOG_INFO
    printf("maxval\t\t%zu\n", maxval);  
    if(Cudd_GarbageCollectionEnabled(dd_mgr)) printf("GC\t\tenabled\n");
    else printf("GC\t\tdisabled\n");    
#endif /*defined(LOG_LEVEL) && LOG_LEVEL <= LOG_INFO*/
    return MY_SUCCESS;
  } else {
#if defined(LOG_LEVEL) && LOG_LEVEL <= LOG_WARN
    fprintf(stderr, "warning: BDD manager already initialized.\n");
#endif /*defined(LOG_LEVEL) && LOG_LEVEL <= LOG_WARN*/
    return MY_FAILURE;
  }
}

static inline int bdd_quit(void)
{
  if(dd_mgr != NULL) {
    Cudd_Quit(dd_mgr);
    dd_mgr = NULL;
    return MY_SUCCESS;
  } else {
#if defined(LOG_LEVEL) && LOG_LEVEL <= LOG_WARN
    fprintf(stderr, "warning: BDD manager already freed.\n");
#endif /*defined(LOG_LEVEL) && LOG_LEVEL <= LOG_WARN*/
    return MY_FAILURE;
  }
}

static inline size_t bdd_label(my_bddp f)
{  
  assert(BDD_NULL != f);
  my_bddp F = Cudd_Regular(f);
  return (size_t)Cudd_NodeReadIndex(F);
}


static inline size_t bdd_size(my_bddp f)
{
  assert(BDD_NULL != f);
  return (size_t)Cudd_DagSize(f);
}


static inline my_bddp bdd_top(void)
{
  assert(dd_mgr != NULL);
  return Cudd_ReadOne(dd_mgr);
}


static inline my_bddp bdd_bot(void)
{
  assert(dd_mgr != NULL);
  return Cudd_ReadLogicZero(dd_mgr);
}

static inline int bdd_isconst(my_bddp f)
{
  assert(BDD_NULL != f);
  return (f==bdd_top() || f==bdd_bot());
}

static inline my_bddp bdd_HI(my_bddp f)
{
  assert(BDD_NULL != f);
  assert(!bdd_isconst(f));

  my_bddp F = Cudd_Regular(f);
  if (Cudd_IsComplement(f)) return Cudd_Not(cuddT(F));
  else return cuddT(F);
}

static inline my_bddp bdd_LO(my_bddp f)
{
  assert(BDD_NULL != f);
  assert(!bdd_isconst(f));

  my_bddp F = Cudd_Regular(f);
  if (Cudd_IsComplement(f)) return Cudd_Not(cuddE(F));
  else return cuddE(F);
}

static inline my_bddp bdd_and(my_bddp f, my_bddp g)
{
  assert(BDD_NULL != f && BDD_NULL != g);
  assert(dd_mgr != NULL);

  return Cudd_bddAnd(dd_mgr, f, g);
}

static inline my_bddp bdd_unique(size_t i, my_bddp lo, my_bddp hi)
{
  assert(i != 0 && i <= BDD_VAR_MAXINDEX);  
  assert(dd_mgr != NULL);
  assert(!bdd_isconst(hi)? i < bdd_label(hi): 1);
  assert(!bdd_isconst(lo)? i < bdd_label(lo): 1);

  my_bddp f;
  if(lo == hi) {
    f = hi;
  } else {
    if (Cudd_IsComplement(hi)) {
      f = cuddUniqueInter(dd_mgr,(int)i,Cudd_Not(hi),Cudd_Not(lo));
      ENSURE_TRUE_MSG(f != BDD_NULL, "BDD operation failed");
      f = Cudd_Not(f);
    } else {
      f = cuddUniqueInter(dd_mgr,(int)i, hi, lo);
      ENSURE_TRUE_MSG(f != BDD_NULL, "BDD operation failed");
    }
  }
  return f;
}


/* Operations for ZDDs */
static inline int zdd_init(size_t maxval, size_t cache_size)
{
  ENSURE_TRUE_MSG(maxval != 0 && maxval <= ZDD_VAR_MAXINDEX && maxval < INFTY, "the number of variables is invalid.");
  if(dd_mgr == NULL) {
    dd_mgr = Cudd_Init((DdHalfWord)maxval, (DdHalfWord)maxval, CUDD_UNIQUE_SLOTS, 16777216, 8589934592);
    //dd_mgr = Cudd_Init((DdHalfWord)maxval, (DdHalfWord)maxval, CUDD_UNIQUE_SLOTS, CUDD_CACHE_SLOTS, 0);
    ENSURE_TRUE_MSG(dd_mgr != NULL, "ZDD manager initialization failed.\n");
    Cudd_DisableGarbageCollection(dd_mgr);// disable GC since we never do dereference of nodes.
#if defined(LOG_LEVEL) && LOG_LEVEL <= LOG_INFO
    printf("maxval\t\t%zu\n", maxval);  
    if(Cudd_GarbageCollectionEnabled(dd_mgr)) printf("GC\t\tenabled\n");
    else printf("GC\t\tdisabled\n");
#endif /*defined(LOG_LEVEL) && LOG_LEVEL <= LOG_INFO*/
    return MY_SUCCESS;
  } else {
#if defined(LOG_LEVEL) && LOG_LEVEL <= LOG_WARN
    fprintf(stderr, "warning: ZDD manager already initialized.\n");
#endif /*defined(LOG_LEVEL) && LOG_LEVEL <= LOG_WARN*/
    return MY_FAILURE;
  }
}

static inline int zdd_quit(void)
{
  if(dd_mgr != NULL) {
    Cudd_Quit(dd_mgr);
    dd_mgr = NULL;
    return MY_SUCCESS;
  } else {
#if defined(LOG_LEVEL) && LOG_LEVEL <= LOG_WARN
    fprintf(stderr, "warning: ZDD manager already freed.\n");
#endif /*defined(LOG_LEVEL) && LOG_LEVEL <= LOG_WARN*/
    return MY_FAILURE;
  }
}

static inline size_t zdd_label(my_zddp f)
{  
  assert(ZDD_NULL != f);
  return (size_t)Cudd_NodeReadIndex(f);
}


static inline size_t zdd_size(my_zddp f)
{
  assert(ZDD_NULL != f);
  return (size_t)Cudd_DagSize(f);
}

static inline my_zddp zdd_bot(void)
{
  assert(dd_mgr != NULL);
  return Cudd_ReadZero(dd_mgr);
}

static inline my_zddp zdd_top(void)
{
  assert(dd_mgr != NULL);
  return Cudd_ReadOne(dd_mgr);
}

static inline int zdd_isconst(my_zddp f)
{
  assert(ZDD_NULL != f);

  return (f == zdd_top()) || (f == zdd_bot());
}

static inline my_zddp zdd_HI(my_zddp f)
{
  assert(ZDD_NULL != f);
  assert(!zdd_isconst(f));

  return cuddT(f);
}

static inline my_zddp zdd_LO(my_zddp f)
{
  assert(ZDD_NULL != f);
  assert(!zdd_isconst(f));

  return cuddE(f);
}

static inline size_t zdd_card(my_zddp f)
{
  assert(ZDD_NULL != f);
  assert(dd_mgr != NULL);
  return (size_t)Cudd_zddCountDouble(dd_mgr, f);
}

static inline my_zddp zdd_union(my_zddp f, my_zddp g)
{
  assert(ZDD_NULL != f && ZDD_NULL != g);
  assert(dd_mgr != NULL);

  return Cudd_zddUnion(dd_mgr, f, g);
}

static inline my_zddp zdd_intersect(my_zddp f, my_zddp g)
{
  assert(ZDD_NULL != f && ZDD_NULL != g);
  assert(dd_mgr != NULL);
  
  return Cudd_zddIntersect(dd_mgr, f, g);
}

static inline my_zddp zdd_diff(my_zddp f, my_zddp g)
{
  assert(ZDD_NULL != f && ZDD_NULL != g);
  assert(dd_mgr != NULL);
  
  return Cudd_zddDiff(dd_mgr, f, g);
}

static inline my_zddp zdd_unique(size_t i, my_zddp lo, my_zddp hi)
{
  assert(i != 0 && i <= ZDD_VAR_MAXINDEX);
  assert(dd_mgr != NULL);
  assert(!zdd_isconst(hi)? i < zdd_label(hi): 1);
  assert(!zdd_isconst(lo)? i < zdd_label(lo): 1);

  return cuddZddGetNode(dd_mgr, (int)i, hi, lo);
}


#else
#error "Specify BDD/ZDD package"
#endif /*defined(SELECT_CUDD)*/

#endif /*DD_WRAPPER_H*/
