/** \file   zdd_extra.c
 *  \brief implementation of ZDD extra operations
 *  \author Takahisa Toda
 */

#include "my_def.h"
#include "my_cache.h"
#include "dd_wrapper.h"
#include "zdd_extra.h"

static struct my_cache *zdd_extra_cache = NULL;

enum ZDD_EXTRA_OPERATIONS {
  ZDD_EXTRA_NONSUP_OP = 1, //!< operation id (a nonzero positive number)
};

static size_t recdepth = 0; //!< recursion depth

/** \brief Preprosessing for ZDD extra operations.
 *  \param cache_size the size of a cache for binary operations
 */
int zdd_init_extra(size_t cache_size)
{
  if(zdd_extra_cache == NULL) {
    zdd_extra_cache = cache_create(cache_size);
    ENSURE_TRUE_MSG(zdd_extra_cache != NULL, "cache creation failed");
    return MY_SUCCESS;
  } else {
#if defined(LOG_LEVEL) && LOG_LEVEL <= LOG_WARN
    fprintf(stderr, "warning: ZDD extra initialization already done.\n");
#endif /*defined(LOG_LEVEL) && LOG_LEVEL <= LOG_WARN*/
    return MY_FAILURE;
  }
}

/** \brief destroy cache for ZDD extra operations
 */
int zdd_quit_extra(void)
{
  if(zdd_extra_cache != NULL) {
    cache_destroy(zdd_extra_cache);
    zdd_extra_cache = NULL;
    return MY_SUCCESS;
  } else {
#if defined(LOG_LEVEL) && LOG_LEVEL <= LOG_WARN
    fprintf(stderr, "warning: ZDD extra cache already freed.\n");
#endif /*defined(LOG_LEVEL) && LOG_LEVEL <= LOG_WARN*/
    return MY_FAILURE;
  }
}

/** \brief Construct the ZDD that represents all sets U such that U in f but U is not superset for any set in g.
 * \param f ZDD
 * \param g ZDD
 * \return the ZDD contructed.
 * \note
 * - See pp.669-670 in "The Art of Computer Programming, Volume 4a (2011)" by Knuth
 */
my_zddp nonsup(my_zddp f, my_zddp g)
{
  if(g == zdd_bot()) return f;
  if(f == zdd_bot() || g == zdd_top() || f == g) return zdd_bot();
  if(f == zdd_top() || zdd_label(f) > zdd_label(g)) return nonsup(f, zdd_LO(g));

  assert(zdd_extra_cache != NULL);
  int res; my_zddp r;
  if((res = cache_search(ZDD_EXTRA_NONSUP_OP, (size_t)f, (size_t)g, (size_t*)&r, zdd_extra_cache)) == MY_SUCCESS) return r;

  INC_RECDEPTH(recdepth);  
  size_t fv = zdd_label(f);
  my_zddp rl; my_zddp rh;

  if(fv < zdd_label(g)) {
    rl = nonsup(zdd_LO(f), g);
    rh = nonsup(zdd_HI(f), g);
  } else {
    rl = nonsup(zdd_HI(f), zdd_HI(g));
    my_zddp t = nonsup(zdd_HI(f), zdd_LO(g));
    rh = zdd_intersect(t, rl);
    ENSURE_TRUE_MSG(rh != ZDD_NULL, "ZDD operation failed");
    rl = nonsup(zdd_LO(f), zdd_LO(g));
  }

  r = zdd_unique(fv, rl, rh);
  ENSURE_TRUE_MSG(r != ZDD_NULL, "ZDD operation failed");

  cache_add(ZDD_EXTRA_NONSUP_OP, (size_t)f, (size_t)g, (size_t)r, zdd_extra_cache);
  DEC_RECDEPTH(recdepth);
  return r;
}


